#include <mpi.h>
#include <omp.h>
#include <iostream>
#include <vector>
#include <functional>
#include <unistd.h>
#include "defs.h"

extern void peer2peer_exchange(const std::vector<int> &params, std::vector<std::vector<double>> &output);
extern void hip_memcopy(const std::vector<int> &params, std::vector<std::vector<double>> &output);
extern int hip_stream(const std::vector<int> &params, std::vector<std::vector<double>> &output);
extern void core_affinity(const std::vector<int> &params, std::vector<std::vector<double>> &output);
extern void nic_affinity(const std::vector<int> &params, std::vector<std::vector<double>> &output);
extern void device_affinity(const std::vector<int> &params, std::vector<std::vector<double>> &output);

void do_test(std::function<void(const std::vector<int> &, std::vector<std::vector<double>> &output)> fn, const std::vector<int> &params, std::vector<std::vector<double>> &output) {
#pragma omp parallel 
    {    
        for (int j = 0; j < 1; j++) {
#pragma omp master
            {
                fn(params, output);
            }
        }
    }
    MPI_Barrier(MPI_COMM_WORLD);
}

struct oglobal {
    std::string header;
    std::vector<double> values;
};

struct olocal {
    std::string header;
    std::vector<std::vector<double>> values;
};

std::string construct_final_output(const std::vector<oglobal> &output) {
//const std::vector<std::vector<double>> &output) {
    std::stringstream ss_outer;
    for (const auto &og : output) {
        ss_outer << ", " << og.header << " {";
        std::stringstream ss_inner;
        ss_inner << std::setprecision(1) << std::fixed;
        for (const auto value : og.values) {
            ss_inner << "," << value;
        }
        ss_outer << ss_inner.str().erase(0, 1) << "}";
    }
	return std::string("{ ") + ss_outer.str().erase(0, 2) + std::string(" }");
}

void local2output(olocal &output_local, std::vector<oglobal> &output) {
    if (output_local.values.empty())
        return;
    std::vector<double> aggregator;
    for (const auto &v : output_local.values) {
        for (const auto value : v) {        
            aggregator.push_back(value);
        }
    }
    output.push_back({output_local.header, aggregator});
    output_local.values.resize(0);
}

int main(int argc, char **argv)
{
    std::vector<oglobal> output;
    MPI_Init(&argc, &argv);
    alarm(170); // NOTE: this alarm is to interrupt test in case of deadlock
    int size, rank;
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    int n = omp_get_max_threads();
    std::cout << ">> rank=" << rank << " omp nthreads=" << n << std::endl;
    MPI_Barrier(MPI_COMM_WORLD);
    do_test(core_affinity, {rank}, output);
    do_test(device_affinity, {rank}, output);
    do_test(nic_affinity, {rank}, output);
    do_test(hip_stream, {rank}, output);
    const int ngpus = 8;
    olocal output_local;
    for (const int use_pinned_memory : {1,0}) { 
        for (int distance = 1; distance < ngpus * 2; distance *= 2) {
            do_test(hip_memcopy, {rank, distance, ngpus, use_pinned_memory}, output_local);
        }
        local2output(output_local, output);
    }
    for (const int use_gpu_memory : {1,0}) { // {1,0} for both GPU and CPU exch tests
        std::string kind = (use_gpu_memory ? "Device2Device" : "Host2Host");
		print(rank, kind + " MPI exchange, intra-node:");
		for (int distance = 1; distance < ngpus; distance *= 2) {
			do_test(peer2peer_exchange, {rank, size, distance, ngpus, use_gpu_memory}, output_local);
		}
        local2output(output_local, output);
        if (size > ngpus) {
            print(rank, kind + "MPI exchange, cross-node:");
            for (int distance = 0; distance < ngpus; distance++) {
                do_test(peer2peer_exchange, {rank, size, distance + ngpus, ngpus, use_gpu_memory}, output_local);
            }
            local2output(output_local, output);
        }
    }
    print(rank, construct_final_output(output));
    MPI_Finalize();
    return 0;
}
